import "Turbine";
import "Turbine.Gameplay";
import "Turbine.UI";
import "Turbine.UI.Lotro";

import "K1R4D3L.KillTracker.Utils";
import "K1R4D3L.KillTracker.UI";

-- identify and set locale
--    beware, if the locale is changed for a character after data is recorded
--    kill counts will not increment correctly if the mobs have different
--    names across the two languages 
_G.locale = "en";

if Turbine.Shell.IsCommand("hilfe") then
	_G.locale = "de";
elseif Turbine.Shell.IsCommand("aide") then
	_G.locale = "fr";
end

import ("K1R4D3L.KillTracker.Locale."..locale);

_G.saveData = true;

_G.isLoaded = false;

hitLog = {}; -- timestamp, initiatorName, targetName
killLog = {}; -- timestamp, initiatorName, targetName, playerLevel
newKills = {}; -- new kill count

isInCombat = false; -- time left combat

-- debugLevels:
--         0 -> quiet, only plugin load/unload and data load failure messages
--         1 -> basic debug info
--         2 ->   + all messages processed
--         3 ->   + hit/kill table dumps with each hit/kill match attempt
_G.debugLevel = 0;

-- show log messages in bright yellow text?
_G.logHightlight = false;

-- send log messages > level 0 to script log instead of chat
_G.logToScriptLog = false;

-- grace period for hit-matching and in-combat checks (seconds)
delaySec = 2;

-- set minimum reasonable window dimensions
_G.minWidth = 520;
_G.minHeight = 200;

-- sort indicators - the unicode characters are nice, but only supported in Arial12
_G.DescSortChar = "+";
_G.AscSortChar = "-"; 
--_G.DescSortChar = string.char(226)..string.char(150)..string.char(188);     
--_G.AscSortChar = string.char(226)..string.char(150)..string.char(178);

_G.options = nil;

-- initialize the plugin
function Init()
   -- get plugin version
   local err, versionNo = pcall(GetVersionNo);
   _G.versionNo = (err and versionNo or "");

   -- get player
   _G.player = Turbine.Gameplay.LocalPlayer.GetInstance();
   player.name = player:GetName();

   -- load window size/position settings
   LoadSettings();
   
   -- create main KillTracker UI window, hidden
   CreateMainWindow();
   
   -- load server labels for same-named, cross-server character differentiation
   LoadServers();
	
   -- load any saved data
   LoadData();
   
   -- load unique server ID for same-named, cross-server character differentiation
   LoadServerID();
   
   InitCharacterNames();
   
   -- register shell command to display main window
   AddShellCommands();

   -- register for chat messages
   AddCallback(Turbine.Chat, "Received", ProcessChatMessages);

   -- register for in-combat state changes
   AddCallback(player, "InCombatChanged", ProcessCombatStateChange);

	-- create event processor timer to check for new kills
	eventProcessor = Timer();
	eventProcessor.OnUpdate = function(sender, timestamp)
		local i = #killLog;
		
		--clean up expired kills, starting at end 
		while (i > 0) do
	      -- toss out kills that have expired without a match
	      if (killLog[i][1] < (timestamp - delaySec)) then
	         for j = 1, i do
				LogMessage("KillTracker: Found no match for \""..killLog[1][2].." killed "..killLog[1][3].."\".", 1, false);
	
				table.remove(killLog, 1);
	         end
	         
	         break;
	      else
	      	i = i - 1;
	      end
		end

		local i = #hitLog;
		
		--clean up expired hits, starting at end 
		while (i > 0) do
	      -- toss out hits that have expired
	      if (hitLog[i][1] < (timestamp - delaySec)) then
	         for j = 1, i do
	         	table.remove(hitLog, 1);
	         end
	         
	         break;
	      else
	      	i = i - 1;
	      end
		end

		if (player:IsInCombat()) then
			eventProcessor.expiresIn = delaySec;			
		end

		if (isLoaded) then
		   CheckForKills();
		end
	end
	eventProcessor.OnExpiration = function()
		if (isLoaded) then
		   CheckForKills();
		end

		isInCombat = false;

		hitLog = {};
		killLog = {};
	end

	-- initialize in-combat state
	ProcessCombatStateChange();

	_G.options = OptionsPanel();

	-- build the options panel
	plugin.GetOptionsPanel = function(self)
		if (options == nil) then
			_G.options = OptionsPanel();
		end
	
		return options;
	end

	-- register unload cleanup function
	plugin.Unload = function(self, sender, args)
		eventProcessor:Stop();

		RemoveShellCommands();
		
		-- remove chat message subscription
		RemoveCallback(Turbine.Chat, "Received", ProcessChatMessages)   
		-- remove in-combat state change subscription
		RemoveCallback(player, "InCombatChanged", ProcessCombatStateChange)   
		
		-- log plugin unload complete
		LogMessage("KillTracker v"..versionNo..L.PluginUnloaded..".", 0, false);
	end

	-- allow data processing to begin
	_G.isLoaded = true;
	
	-- log plugin startup complete
	LogMessage("KillTracker v"..versionNo..L.PluginLoaded..".", 0, false);
end

function ProcessCombatStateChange()
	if (player:IsInCombat()) then
		isInCombat = true;
		
		eventProcessor.expiresIn = delaySec;
		eventProcessor:Start();
	end
end

function ProcessChatMessages(sender, args)
	if (isLoaded) then
	   -- only process Combat and Death channels
		if ((args.ChatType ~= Turbine.ChatType.EnemyCombat) and (args.ChatType ~= Turbine.ChatType.PlayerCombat) and (args.ChatType ~= Turbine.ChatType.Death)) then
			return;
		end
		
		local timestamp = Turbine.Engine.GetLocalTime();
		
	   -- strip formatting from message and parse using localized parser 
		local updateType, initiatorName, targetName = L.Parse(string.gsub(string.gsub(args.Message, "<rgb=#......>(.*)</rgb>", "%1"), "^%s*(.-)%s*$", "%1"), args.ChatType);
	
		if (updateType == nil) then
	      return;
	   end
		
	   local tmpMsg = "";
	   
	   -- a mob hit or reflect, add to hit log
		if (updateType == 1) then
	      tmpMsg = initiatorName.." hit "..targetName;
	
	      table.insert(hitLog, {timestamp, initiatorName, targetName});
	   -- a mob kill
		elseif (updateType == 9) then
	   	-- grab the character's current level for the log
	   	local playerLevel = player:GetLevel();
	
	      -- direct kill by character, credit the character with the kill
	      if (initiatorName == "$PLAYER$") then
	         UpdateKillsTable(targetName, playerLevel, timestamp);
	         
	         LogMessage("KillTracker: "..player.name.." killed "..targetName..".", 1, false);
	      -- not a direct kill by character, add to kill log
	      else 
	         tmpMsg = initiatorName.." killed "..targetName;
	
	         table.insert(killLog, {timestamp, initiatorName, targetName, playerLevel});
	      end
		end
		
	   -- log logic result and original message
		LogMessage("KillTracker: "..tmpMsg..","..timestamp..","..args.Message, 2, false);
	end
end

-- attempt to ID kills to be credited to character by comparing kill log with hit and in-combat state
function CheckForKills()
   -- dump table contents to log
	if (debugLevel > 2) then
      if (#killLog > 0) then
			for x = 1, #killLog do
         	LogMessage("KillTracker: kill:"..killLog[x][1]..","..killLog[x][2]..","..killLog[x][3], 3, false);
      	end
      end

      if (#hitLog > 0) then
			for x = 1, #hitLog do
         	LogMessage("KillTracker: hit:"..hitLog[x][1]..","..hitLog[x][2]..","..hitLog[x][3], 3, false);
      	end
      end
   end

   local i = 1;

   -- loop through kills
   while ((#killLog > 0) and (i <= #killLog)) do
		-- if the player was in combat during the kill, check for a matching hit
		if (isInCombat) then
			local k = 1;
		
			-- loop through hit log
			while ((#hitLog > 0) and (k <= #hitLog)) do
		      -- look for hit with same initiator and target as kill
		      if ((killLog[i][2] == hitLog[k][2]) and (killLog[i][2] == hitLog[k][2])) then
		         -- credit the character with the kill
		         UpdateKillsTable(killLog[i][3], killLog[i][4], killLog[i][1]);
		         
		         LogMessage("KillTracker: "..killLog[i][2].." killed "..killLog[i][3]..".", 1, false);
		
		         -- remove the matching rows from the logs and move to next kill
		         table.remove(killLog, i);
		         table.remove(hitLog, k);
		
		         break;
		      end
		
		      k = k + 1;
			end
			
			i = i + 1;
		-- player wasn't in combat during kill, don't give credit
		else
		  LogMessage("KillTracker: Player not in combat during \""..killLog[i][2].." killed "..killLog[i][3].."\".", 1, false);
		
		  -- toss out the uncredited kill
		  table.remove(killLog, i);
		end
   end
end

-- record newly-credited kills to master kill log 
function UpdateKillsTable(targetName, playerLevel, timestamp)
	local rowIndex = nil; 

   -- need to add a new entry for this target for this character   
	if (newKills[targetName] == nil) then
		for i = 1, #killsTable do
			if ((killsTable[i][5] == player.uniqueName) and (killsTable[i][1] == targetName)) then
				rowIndex = i;
				
				newKills[targetName] = {1, i};
			
				break;
			end
		end

		if (rowIndex == nil) then
			table.insert(killsTable, {targetName, 1, playerLevel, playerLevel, player.uniqueName, timestamp});
	
			newKills[targetName] = {1, #killsTable};
		end
	else
		rowIndex = newKills[targetName][2];
	
		newKills[targetName][1] = newKills[targetName][1] + 1;
	end  

	if (rowIndex ~= nil) then
		-- increment kill count for this target
		killsTable[rowIndex][2] = killsTable[rowIndex][2] + 1;

		-- update character level at latest kill for this target
		killsTable[rowIndex][4] = playerLevel;
		
		-- update timestamp at latest kill for this target
		killsTable[rowIndex][6] = timestamp;
	end
   
	if (debugLevel > 2) then
		for k, v in pairs(newKills) do
       	LogMessage("KillTracker: newKills:"..k..":"..v[1].."|"..v[2], 3, false);
    	end
	end

   -- save data to Account-level file
   SaveData();

   -- refresh the main KillTracker window
   RefreshKillsDisplay();
end

-- build character name lookup table for display, removing the ServerID part if
--    that character name is unique in the data, including the current player
function InitCharacterNames()
	_G.characterNames = {};

	local characterShortNames = {}; 

	for i = 1, #killsTable do
		-- if this is a new characterName + ServerID
		if (characterNames[killsTable[i][5]] == nil) then
			-- if the shortened name (just characterName, no Server ID) has not been used yet
			--    set the display name for the character to the shortened name
			if (characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)] == nil) then
				characterNames[killsTable[i][5]] = string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1);
			-- there are duplicate character names, switch them both to display the long names
			else
				characterNames[killsTable[i][5]] = killsTable[i][5];
				characterNames[characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)]] = characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)];
			end
				
			-- store the full name under the short name key in case we later find a duplicate
			characterShortNames[string.sub(killsTable[i][5], 1, string.find(killsTable[i][5], "(", 1, true) - 1)] = killsTable[i][5];
		end
	end
	
	-- add the current player if they're not already in the data
	if (characterNames[player.uniqueName] == nil) then
		if (characterShortNames[player.name] == nil) then
			characterNames[player.uniqueName] = player.name;
		else
			characterNames[player.uniqueName] = player.uniqueName;
			characterNames[characterShortNames[player.name]] = characterShortNames[player.name];
		end
	end
end

-- initialize plugin
Init();
